/*========================================================================
     "gpop1scf"  Gaussian output post processor 1
        - read Gaussian SCF-freq job output and create a .gpo file

             Copyright (c) 2002-2006 by A. Miyoshi, Univ. Tokyo
                            GauProc1 created: Oct.  12, 2002
                                      edited: Apr.  26, 2003
                            gpop1scf created: Sept.  1, 2004
                                      edited: Jan.  13, 2005
                                      edited: Aug.   2, 2005
                                    (revised to support #T output)
                                      edited: Aug.  22, 2006
--------------------------------------------------------------------------
 usage: gpop1scf basename[.ext] [-csv/-all]
--------------------------------------------------------------------------
 - Compatible with output of Gaussian 94, 98, and 03.
 - If no extension, default extension '.log' will be added.
 - By default, files basename.gpo and basename_.mod is created,
    while with '-csv' option, 'basename.csv' is also created.
    With '-all' option, .gpo, .csv, .frq, .geo, .tst, and .zpe files
     are created.
--------------------------------------------------------------------------
 [examples]
 % gpop1scf methane
  - methane.log is read, methane.gpo is created.
 % gpop1scf ethane.out -csv
  - ethane.out is read, ethane.gpo and ethane.scv are created.
========================================================================*/

#include "gpop1scf.h"

/*===== main function ==================================================*/

int main(int argc, char *argv[]) {
  int ifrq, irot, flgocsv = false, flgoall = false, badopt = false;
  string gaufln, opt, basfln, extfln, route, comm, gauv, esta, scfe, chml;
  ifstream inf;
  vector<string> geom, freq;
  jobInfos jbinf;
  molecGeom molc;
  freqTable frtab;

  if (argc > 1) { gaufln = argv[1]; }
  if (argc > 2) {
    opt = argv[2];
    if (opt == "-csv") { flgocsv = true; }
    else if (opt == "-all") { flgoall = true; }
    else { badopt = true; }
  }
  if ((argc < 2) || (argc > 3) || ((argc == 3) && badopt))
    { cout << "gpop1scf basename[.ext] [-csv/-all]\n"; return 0; }
   basfln = getBasePart(gaufln); extfln = getExtPart(gaufln);
   if (extfln == "") { gaufln = basfln + ".log"; }

  // --- read in gaussian output
  inf.open(gaufln.c_str());
  if (!inf) { cout << "failed to open [" << gaufln << "].\n"; return 0; }
  if (findFreqJob(inf, gauv, route, comm) == ERRE)
    { cout << "no frequency job was found.\n"; return 0; }
  if (getGeoFrq(inf, geom, freq, esta, scfe, chml) == ERRE)
    { cout << "no frequency output was found.\n"; return 0; }
  inf.close();

  // --- process job-infos, molec-grom and freq-table
  jbinf.setFrom(gauv, route, comm, scfe);
  if (molc.setFrom(geom, esta, chml) == ERRE)
    { cout << "geom read error.\n"; return 0; }
  if (frtab.setFrom(freq, jbinf, molc) == ERRE)
    { cout << "freq read error.\n"; return 0; }

  // --- output .geo, .frq, .csv, .tst, .zpe, .gpo, and _.mod
  if (flgoall) {
    if (outFileGeo(basfln, geom) == ERRE) { return 0; }
    if (outFileFrq(basfln, freq, jbinf, molc) == ERRE) { return 0; }
    if (outFileTst(basfln, jbinf, molc, frtab) == ERRE)
      { return 0; }
    if (outFileZpe(basfln, jbinf, frtab) == ERRE) { return 0; }
  }
  if (flgoall || flgocsv) {
    if (outFileCsv(basfln, jbinf, molc, frtab) == ERRE) { return 0; }
  }
  if (outFileGpo(basfln, jbinf, molc, frtab, gaufln) == ERRE)
    { return 0; }
  if (outFileModTemplate(basfln, jbinf, molc, frtab) == ERRE)
    { return 0; }

  return 0;
}

/*===== find frequency job =============================================*/

int findFreqJob(istream &inf, string &gauv, string &route, string &comm) {
  int found = false;
  do {
    if (findJobHead(inf, gauv) == ERRE) { return ERRE; }
    if ((route = getRouteSect(inf)) == "") { return ERRE; }
    if (route.find("FREQ") != string::npos) { found = true; }
    else { if (findJobEnd(inf) == ERRE) { return ERRE; } }
  } while (!found);
  comm = getCommLine(inf);
  return NORM;
}

/*===== get geometry and frequencies parts =============================*/

int getGeoFrq(istream &inf, vector<string> &gm, vector<string> &fr,
 string &estate, string &scfene, string &chgmul) {
  int freqfound = false, geomend, freqend;
  string slin;
  do {
    if (getLine(inf, slin) == ERRE) { return ERRE; }
    if (slin.substr(0, 16) == " Framework group") {
      gm.clear(); gm.push_back(slin); geomend = false;
      do {
        if (getLine(inf, slin) == ERRE) { return ERRE; }
        if (slin.substr(0, 21) == " Rotational constants")
          { geomend = true; }
        gm.push_back(slin);
      } while (!geomend);
    }
    if (slin.substr(0, 21) == " Harmonic frequencies") {
      do {
        if (getLine(inf, slin) == ERRE) { return ERRE; }
      } while (slin.rfind("normal coordinates:") == string::npos);
      fr.clear(); freqend = false;
      do {
        if (getLine(inf, slin) == ERRE) { return ERRE; }
        if (slin.length() < 22) { freqend = true; }
        else { fr.push_back(slin); }
      } while (!freqend);
      freqfound = true;
    }
    if (slin.find("Multiplicity") != string::npos) { chgmul = slin; }
    if (slin.find("electronic state") != string::npos) { estate = slin; }
    if (slin.find("SCF Done:") != string::npos) { scfene = slin; }
  } while (!freqfound);
  return NORM;
}

/*----- find gaussian job head -----------------------------------------*/

int findJobHead(istream &inf, string &gauv) {
  int found = false;
  string slin;
  do {
    if (getLine(inf, slin) == ERRE) { return ERRE; }
    if (slin.substr(0, 9) == " ********") {
      if (getLine(inf, slin) == ERRE) { return ERRE; }
      if (slin.substr(0, 9) == " Gaussian")
        { gauv = slin; found = true; }
    } else if (slin.substr(0, 19) == " Link1:  Proceeding") {
      found = true;
    }
  } while (!found);
  return NORM;
}

/*----- get route section ----------------------------------------------*/

string getRouteSect(istream &inf) {
  int endroute = false;
  string slin, route;
  do {
    if (getLine(inf, slin) == ERRE) { return ""; }
    if (slin.substr(0, 6) == " -----") {
      if (getLine(inf, slin) == ERRE) { return ""; }
      if (slin.substr(0, 2) == " #") {
        route = slin;
        do {
          if (getLine(inf, slin) == ERRE) { return ""; }
          if (slin.substr(0, 6) == " -----") { endroute = true; }
          else { route += slin; }
        } while (!endroute);
      }
    }
  } while (!endroute);
  strAllToUpper(route);
  return route;
}

/*----- get comment line -----------------------------------------------*/

string getCommLine(istream &inf) {
  int endcomm = false;
  string slin, comm;
  do {
    if (getLine(inf, slin) == ERRE) { return ""; }
    if (slin.substr(0, 3) == " --") {
      if (getLine(inf, slin) == ERRE) { return ""; }
      comm = slin;
      do {
        if (getLine(inf, slin) == ERRE) { return ""; }
        if (slin.substr(0, 3) == " --") { endcomm = true; }
        else { comm += slin; }
      } while (!endcomm);
    }
  } while (!endcomm);
  if (comm.length() > 0) {
    if (comm[0] == ' ') { comm = comm.substr(1); }
  }
  return comm;
}

/*----- find gaussian job end ------------------------------------------*/

int findJobEnd(istream &inf) {
  int found = false;
  string slin;
  do {
    if (getLine(inf, slin) == ERRE) { return ERRE; }
    if (slin.substr(0, 14) == " Job cpu time:") {
      if (getLine(inf, slin) == ERRE) { return ERRE; }
      if (getLine(inf, slin) == ERRE) { return ERRE; }
      if (slin.find("termination") != string::npos) { found = true; }
    }
  } while (!found);
  return NORM;
}

/*===== output .geo file ===============================================*/

int outFileGeo(string &basfln, vector<string> &geom) {
  string goutfln;
  ofstream gof;
  vector<string>::iterator ps;

  goutfln = basfln + ".geo";
  gof.open(goutfln.c_str());
  if (!gof)
    { cout << "failed to open [" << goutfln << "].\n"; return ERRE; }
  for (ps = geom.begin(); ps != geom.end(); ps++) { gof << *ps << endl; }
  gof.close();
  return NORM;
}

/*===== output .frq file ===============================================*/

int outFileFrq(string &basfln, vector<string> &freq, jobInfos &ji,
 molecGeom &mg) {
  string foutfln;
  ofstream fof;
  vector<string>::iterator ps;

  foutfln = basfln + ".frq";
  fof.open(foutfln.c_str());
  if (!fof)
    { cout << "failed to open [" << foutfln << "].\n"; return ERRE; }
  fof << ji.comment << endl;
  fof << mg.nAtms << endl;
  fof << mg.nVibm << endl;
  for (ps = freq.begin(); ps != freq.end(); ps++) { fof << *ps << endl; }
  fof.close();
  return NORM;
}

/*===== output .csv file ===============================================*/

int outFileCsv(string &basfln, jobInfos &ji, molecGeom &mg,
 freqTable &ft) {
  int ifrq, prec, irot, iatm, ixyz;
  string csvfln;
  ofstream cof;
  map<int, int>::iterator p;

  csvfln = basfln + ".csv";
  cof.open(csvfln.c_str());
  if (!cof)
    { cout << "failed to open [" << csvfln << "].\n"; return ERRE; }

  cof << "[Job informations]\n";
  cof << "version,Gaussian " << ji.version << endl;
  cof << "revision," << ji.revision << endl;
  cof << "comment," << ji.comment << endl;
  cof << "methodIdx," << ji.methodIdx << endl;
  prec = cof.precision(); cof.precision(12);
  cof << "scaleFactVib," << ji.scaleFactVib << endl;
  cof << "scaleFactZPE," << ji.scaleFactZPE << endl;
  cof << "scaleFactGT," << ji.scaleFactGT << endl;
  cof.precision(prec);
  cof << "SCFenergy," << ji.SCFenergy << endl;
  cof << endl;

  cof << "[Geometry and related properties]\n";
  cof << "nAtms," << mg.nAtms << endl;
  cof << "gElec," << mg.gElec << endl;
  cof << "nVibm," << mg.nVibm << endl;
  cof << "isLinear,";
  if (mg.isLinear) { cof << "true\n"; }
  else { cof << "false\n"; }
  cof << "pntGrp," << mg.pntGrp << endl;
  cof << "elState," << mg.elState << endl;
  cof << "rotSymNbr," << mg.rotSymNbr << endl;
  cof << "numIsomers," << mg.numIsomers << endl;
  prec = cof.precision(); cof.setf(ios::fixed); cof.precision(6);
  cof << "molW," << mg.molW << endl;
  cof.precision(7);
  cof << "rotcGHz";
  for (irot = 0; irot < 3; irot++) { cof << ',' << mg.rotcGHz[irot]; }
  cof << endl;
  cof << "chemform";
  for (p = mg.chemform.begin(); p != mg.chemform.end(); p++)
    { cof << ',' << elm.getAtmcSym(p->first) << ',' << p->second; }
  cof << endl;
  cof.precision(6);
  cof << "centNbr,atmcNbr,atmcTyp,X,Y,Z\n";
  for (iatm = 0; iatm < mg.nAtms; iatm++) {
    cof << iatm + 1 << ',' << mg.geom[iatm].atomicNumber << ','
        << mg.geom[iatm].atomicType;
    for (ixyz = 0; ixyz < 3; ixyz++)
      { cof << ',' << mg.geom[iatm][ixyz]; }
    cof << endl;
  }
  cof.unsetf(ios::fixed); cof.precision(prec);
  cof << endl;

  cof << "[Vibrational frequencies and related properties]\n";
  prec = cof.precision(); cof.setf(ios::fixed); cof.precision(4);
  cof << "zpe_Unscaled," << ft.zpe_Unscaled << endl;
  cof << "zpe," << ft.zpe << endl;
  cof << "zpe_GT," << ft.zpe_GT << endl;
  cof << "vibNbr,sym,freq,rMass,fConst,IRint,ramAct,depol\n";
  for (ifrq = 0; ifrq < mg.nVibm; ifrq++) {
    cof << ifrq + 1 << "," << ft[ifrq].sym << ",";
    cof.setf(ios::fixed); cof.precision(4);
    cof << ft[ifrq].freq << "," << ft[ifrq].rmass << ","
      << ft[ifrq].fconst << "," << ft[ifrq].irint << ","
      << ft[ifrq].ramact << "," << ft[ifrq].depol << endl;
  }
  cof << "vibNbr,centNbr,dX,dY,dZ\n";
  cof.precision(2);
  for (ifrq = 0; ifrq < mg.nVibm; ifrq++) {
    cof << ifrq + 1 << endl;
    for (iatm = 0; iatm < mg.nAtms; iatm++) {
      cof << ',' << iatm + 1;
      for (ixyz = 0; ixyz < 3; ixyz++)
        { cof << ',' << ft[ifrq].nrm[iatm][ixyz]; }
      cof << endl;
    }
  }
  cof.unsetf(ios::fixed); cof.precision(prec);

  cof.close();
  return NORM;
}

/*===== output .zpe file ===============================================*/

int outFileZpe(string &basfln, jobInfos &ji, freqTable &ft) {
  string zoutfln;
  ofstream zof;

  zoutfln = basfln + ".zpe";
  zof.open(zoutfln.c_str());
  if (!zof)
    { cout << "failed to open [" << zoutfln << "].\n"; return ERRE; }
  zof << "method detected as " << ji.method << "/" << ji.basis << ".\n";
  zof << "scaling factors for Vib.Freq. / ZPE / Gau.Theory :\n";
  zof << " " << ji.scaleFactVib << " / " << ji.scaleFactZPE << " / "
      << ji.scaleFactGT << endl;
  zof.setf(ios::fixed); zof.precision(4);
  zof << "- ZPE(unscaled)  ZPE  ZPE(for Gaussian Theory) /cm-1\n";
  zof << " " << ft.zpe_Unscaled << "  " << ft.zpe << "  " << ft.zpe_GT
      << endl;
  zof.precision(10);
  zof << "- ZPE(unscaled)  ZPE  ZPE(for Gaussian Theory) /hartree\n";
  zof << " " << ft.zpe_Unscaled / WAVNHRT << "  " << ft.zpe / WAVNHRT
      << "  " << ft.zpe_GT / WAVNHRT << endl;
  zof.close();
  return NORM;
}

/*===== output .mod template file ======================================*/

int outFileModTemplate(string &basfln, jobInfos &ji, molecGeom &mg,
 freqTable &ft) {
  int isTS = false;
  string modfln;
  ofstream mof;

  modfln = basfln + "_.mod";
  mof.open(modfln.c_str());
  if (!mof)
    { cout << "failed to open [" << modfln << "].\n"; return ERRE; }
  mof << "! gElec " << mg.gElec << endl;
  mof << "! rotSymNbr " << mg.rotSymNbr << endl;
  mof << "! numIsomers " << mg.numIsomers << endl;
  if (ft[0].freq < 0.) { isTS = true; }
  mof << "! isTS " << strGpoBlnv(isTS) << endl;
  mof << "! energyTST #\n";
  mof << "! setIntRotor idVib nSym moi1 moi2 comment\n";
  mof.close();
  return NORM;
}

